from handlerworkers.apikeys import CSAPIKeys
from handlerworkers.cluster import add_cluster_command, get_app_keys, get_appliance_keys, get_postman_keys
from handlerworkers.cluster import save_appliance_keys_for_new_nodes, save_postman_keys_for_new_nodes, save_app_keys_for_new_nodes, save_encryption_key_for_new_nodes

from os import makedirs
import argparse
from os.path import join,isfile
from utilities.util import write_to_file
import sys
import random, string


def generate_appliance_keys(propagate_to_cluster = True):
    keys_dir = '/opt/cyops/configs/keys/'
    makedirs(keys_dir, exist_ok=True)

    if propagate_to_cluster:
        appliance_private_key, appliance_public_key = CSAPIKeys().refresh_keys(uuid='6f3626c1-0da7-4b44-8d5e-43750f780a47')
    else:
        appliance_keys = get_appliance_keys()
        appliance_private_key, appliance_public_key = appliance_keys['private_key'], appliance_keys['public_key']

    write_to_file(appliance_public_key, join(keys_dir, 'APPLIANCE_PUBLIC_KEY'))
    write_to_file(appliance_private_key, join(keys_dir, 'APPLIANCE_PRIVATE_KEY'))

    postman_enabled = True
    try:
        if propagate_to_cluster:
            postman_private_key, postman_public_key = CSAPIKeys().refresh_postman_keys()
        else:
            postman_keys = get_postman_keys()
            postman_private_key, postman_public_key = postman_keys['private_key'], postman_keys['public_key']
        write_to_file(postman_public_key, join(keys_dir, 'POSTMAN_APPLIANCE_PUBLIC_KEY'))
        write_to_file(postman_private_key, join(keys_dir, 'POSTMAN_APPLIANCE_PRIVATE_KEY'))
    except Exception:
        # postman keys are not present. put appliance keys only
        print('postman key not found. This is not a distributed deployment')
        postman_enabled = False

    if propagate_to_cluster:
        appliance_keys = {}
        appliance_keys['public_key'], appliance_keys['private_key'] = appliance_public_key, appliance_private_key
        save_appliance_keys_for_new_nodes(appliance_keys)
        if(postman_enabled):
            postman_keys = {}
            postman_keys['public_key'], postman_keys['private_key'] = postman_public_key, postman_private_key
            save_postman_keys_for_new_nodes(postman_keys)
        add_cluster_command('generate-appliance-keys', '/opt/cyops-auth/.env/bin/python /opt/cyops-auth/defaultappliance.py --no-propagate generate-appliance-keys')


def generate_app_keys(propagate_to_cluster, app=None):
    keys_dir = '/etc/pki/cyops/'
    if not app:
        keys_to_generate = ["das", "wf", "jwt", "postman", "integrations"]
    else:
        keys_to_generate = [app]

    makedirs(keys_dir, exist_ok=True)

    if propagate_to_cluster:
        app_keys = {}
        for key_type in keys_to_generate:
            private_key, public_key = CSAPIKeys().generate_key_pairs()
            write_to_file(public_key, join(keys_dir, key_type+'public.key'))
            write_to_file(private_key, join(keys_dir, key_type+'private.key'))
            app_keys[key_type] = {}
            app_keys[key_type]['public_key'], app_keys[key_type]['private_key'] = public_key, private_key
        save_app_keys_for_new_nodes(app_keys)
        add_cluster_command('generate-app-keys',
                            '/opt/cyops-auth/.env/bin/python /opt/cyops-auth/defaultappliance.py --no-propagate generate-app-keys')
    else:
        app_keys = get_app_keys()
        for key_type in keys_to_generate:
            private_key, public_key = app_keys[key_type]['private_key'], app_keys[key_type]['public_key']
            write_to_file(public_key, join(keys_dir, key_type + 'public.key'))
            write_to_file(private_key, join(keys_dir, key_type + 'private.key'))


def generate_encryption_key(propagate_to_cluster):
    key_path = '/opt/cyops/configs/cyops-api/application.conf'
    if not isfile(key_path):
        encryption_key = 'ssl_key='+''.join([random.choice(string.ascii_letters + string.digits) for n in range(32)])
        write_to_file(encryption_key, key_path)
        if propagate_to_cluster:
            save_encryption_key_for_new_nodes(encryption_key)


def take_arguments():
    parser = argparse.ArgumentParser(description='Utility to generate appliance and app keys')

    subparsers = parser.add_subparsers(help='commands')

    appliance_keys_parser = subparsers.add_parser('generate-appliance-keys', help='Generate Appliance Keys')
    appliance_keys_parser.set_defaults(func=generate_appliance_keys)

    app_keys_parser = subparsers.add_parser('generate-app-keys', help='Generate App Keys i.e. das, jwt, wf and postman')
    app_keys_parser.set_defaults(func=generate_app_keys)

    app_keys_parser = subparsers.add_parser('generate-encryption-key', help='Generate Database Encryption Key')
    app_keys_parser.set_defaults(func=generate_encryption_key)

    parser.add_argument('--no-propagate', dest='propagate_to_cluster', action = 'store_false', help = 'Do no propagate the command to cluster')
    parser.add_argument('--app', dest='app', action = 'store', help = 'Any specific app to generate keys')

    args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])
    return args

if __name__ == '__main__':
    args = take_arguments()

    if hasattr(args, 'func'):
        if args.func.__name__ == 'generate_app_keys':
            args.func(args.propagate_to_cluster, args.app)
        else:
            args.func(args.propagate_to_cluster)
