#!/opt/cyops-auth/.env/bin/python
import argparse, sys, socket, json, os
import signal

if os.getuid() != 0:
    print("This script must be run as 'root'")
    exit(1)

def signal_handler(signal, frame):
    sys.exit(1)

signal.signal(signal.SIGINT, signal_handler)

sys.path.append("/opt/cyops-auth")
from utilities.ha.conf import read_config_file
from utilities.ha.handler import CyOpsHAHandler
from utilities.ha.message_constants import *
from utilities.csconstants import CSConstants
from utilities.ha.common_utils import validate_file, print_dataset, CustomHelpFormatter, print_error
from handlerworkers.cluster import add_sync_cluster_command, get_cluster_commands, add_cluster_command, add_cluster_publish
from utilities.ha.consts import CONF_FILE, POSTGRES_SECTION
from utilities.ha.postgres import verify_postgresql_credentials
from log.cslog import get_ha_logger
logger = get_ha_logger(__name__)


def add_joincluster_subparser(subparsers):
    status_choices = ['active', 'passive']
    role_choices = ['primary', 'secondary']
    parser_join_cluster = subparsers.add_parser('join-cluster', help='Join the HA cluster')
    parser_join_cluster.set_defaults(operation='join-cluster')

    parser_join_cluster.add_argument('--status', action='store', type=str, choices=status_choices, required=True)

    # --role required only in case of --status 'active'
    parser_join_cluster.add_argument('--role', action='store', type=str, choices=role_choices,
                                     required='--status' in sys.argv and 'active' in sys.argv,
                                     default='primary',
                                     help='Required only if status is "active"')

    parser_join_cluster.add_argument('--conf', action='store', type=str, dest='config_file_path',
                                     default=None,
                                     help='Conf file copied from primary server')

    parser_join_cluster.add_argument('--primary-node', action='store', type=str, dest='primary_node_hostname',
                                     default=None,
                                     help='Primary node DNS name')

    parser_join_cluster.add_argument('--primary-node-ssh-key', action='store', type=str, dest='primary_node_ssh_key_path',
                                     default=None,
                                     help='SSH private key of primary node')

    parser_join_cluster.add_argument('--fetch-fresh-backup', action='store_true', dest='fetchfreshbackup', help='Take a fresh backup from primary')
    
    # Node name will be default to hostname
    curr_hostname = socket.gethostname()
    parser_join_cluster.add_argument('--node_name', action='store', type=str, nargs='?', const=1, default=curr_hostname,
                                     help='By default set as the current hostname')


def add_get_config_subparser(subparsers):
    parser_getconf = subparsers.add_parser('export-conf', help='Export the configuration file')
    parser_getconf.set_defaults(operation='export-conf')

def add_whitelist_subparser(subparsers):
    parser_register_server = subparsers.add_parser('whitelist', help="Whitelist secondary/passive server")
    parser_register_server.set_defaults(operation="whitelist")
    parser_register_server.add_argument("--nodes", action='store', type=str, help="Comma-separated list of hostnames of the cluster nodes", required=True)


def add_list_nodes_subparser(subparsers):
    parser_listnodes = subparsers.add_parser('list-nodes', help="List HA cluster details")
    parser_listnodes.set_defaults(operation="list-nodes")
    parser_listnodes.add_argument('--active', action='store_true', dest='is_active', default=None, help='List all active nodes')
    parser_listnodes.add_argument('--passive', action='store_false', dest='is_active', default=None, help='List all passive nodes')
    parser_listnodes.add_argument('--primary', action='store_true', dest='is_primary', default=None, help='List all primary nodes in cluster')
    parser_listnodes.add_argument('--secondary', action='store_false', dest='is_primary', default=None, help='List all secondary nodes in cluster')


def add_leave_cluster_subparser(subparsers):
    parser_leavecluster = subparsers.add_parser('leave-cluster', help="Leave HA cluster")
    parser_leavecluster.set_defaults(operation="leave-cluster")
    parser_leavecluster.add_argument('--no-interaction', action='store_false', dest='interactive', default=True,
                                     help="Don't ask for any confirmation")

def add_list_command_subparser(subparsers):
    listcommands = subparsers.add_parser('list-commands', help='List pending cluster commands')
    listcommands.set_defaults(func=listcommand_impl, operation='list-commands')
    listcommands.add_argument('--status', action='store', type=str, required=False)
    listcommands.add_argument('--nodeId', action='store', type=str, required=False)

def add_add_command_subparser(subparsers):
    addcommandsync = subparsers.add_parser('add-command')
    addcommandsync.set_defaults(func=addcommand_impl, operation='add-command')
    addcommandsync.add_argument('--async', action='store_true', required=False)
    addcommandsync.add_argument('--commandId', action='store', type=str, required=True)
    addcommandsync.add_argument('--command', action='store', type=str, required=True)
    addcommandsync.add_argument('--nodeId', action='store', type=str, required='--async' not in sys.argv)
    addcommandsync.add_argument('--timeout_in_mins', action='store', type=int, required='--async' not in sys.argv, default=2)

def add_add_message_subparser(subparsers):
    addmessage = subparsers.add_parser('add-message')
    addmessage.set_defaults(func=addmessage_impl, operation='add-message')
    addmessage.add_argument('--message', action='store', type=str, required=True)
    addmessage.add_argument('--route', action='store', type=str, required=True)
    addmessage.add_argument('--active', action='store_true', required=False)

def add_update_hostname_subparser(subparsers):
    update_hostname_parser = subparsers.add_parser('set-node-name')
    update_hostname_parser.set_defaults(operation='set-node-name')
    update_hostname_parser.add_argument("node_name", action='store', type=str)


def define_parser(parser):
    visible_options = "{join-cluster,export-conf,whitelist,list-nodes,leave-cluster,list-commands,takeover,firedrill,restore}"
    subparsers = parser.add_subparsers(metavar=visible_options)

    add_joincluster_subparser(subparsers)
    add_get_config_subparser(subparsers)
    add_whitelist_subparser(subparsers)
    add_list_nodes_subparser(subparsers)
    add_leave_cluster_subparser(subparsers)
    add_list_command_subparser(subparsers)
    add_add_command_subparser(subparsers)
    add_add_message_subparser(subparsers)
    add_update_hostname_subparser(subparsers)

    parser_takeover = subparsers.add_parser('takeover', help="Perform takeover")
    parser_takeover.set_defaults(operation="takeover")
    parser_takeover.add_argument('--no-interaction', action='store_false', dest='interactive', default=True,
                                 help="Don't ask for any confirmation")

    parser_firedrill = subparsers.add_parser('firedrill', help="Test DR")
    parser_firedrill.set_defaults(operation="firedrill")

    parser_restore = subparsers.add_parser('restore', help="Restore the server to either passive/secondary after firedrill")
    parser_restore.set_defaults(operation="restore")

    for name, subp in subparsers.choices.items():
         subp._positionals.title = 'subcommand options are'
         subp._optionals.title = 'subcommand options are'

def listcommand_impl(*args, **kwargs):
    command_args = args[0]
    config = read_config_file(CONF_FILE)
    verify_postgresql_credentials(config.get(POSTGRES_SECTION), True)
    commands = get_cluster_commands(command_args['status'], command_args['nodeId'])
    print_dataset(commands)


def addcommand_impl(*args, **kwargs):
    command_args = args[0]
    is_async = command_args['async']
    if is_async:
        add_cluster_command(command_args['commandId'], command_args['command'], command_args['nodeId'])
    else:
        print(json.dumps(add_sync_cluster_command(command_args['commandId'], command_args['command'],
                                              command_args['nodeId'], command_args['timeout_in_mins'])))


def addmessage_impl(*args, **kwargs):
    command_args = args[0]
    if command_args['active']:
        print(json.dumps(add_cluster_publish(command_args['message'], command_args['route'],
                                         is_active=command_args['active'])))
    else:
        print(json.dumps(add_cluster_publish(command_args['message'], command_args['route'])))

def validate_joincluster_inputs(args):
    if args.primary_node_hostname is None and args.config_file_path is None:
        print_error(mconsts_9)
        exit(1)

def start():

    parser = argparse.ArgumentParser(prog="csadm ha", formatter_class=CustomHelpFormatter)
    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'
    define_parser(parser)

    args = parser.parse_args()

    if not 'operation' in dir(args):
        parser.print_help(sys.stderr)
        exit(1)

    if 'config_file_path' in dir(args) and args.config_file_path is not None:
        config_file_path = args.config_file_path.strip()
        validate_file(config_file_path)

    if 'primary_node_ssh_key_path' in dir(args) and args.primary_node_ssh_key_path is not None:
        primary_node_ssh_key_path = args.primary_node_ssh_key_path.strip()
        validate_file(primary_node_ssh_key_path)

    if 'func' in dir(args):
        args.func(vars(args))

    if str(args.operation) == 'join-cluster':

        handler = CyOpsHAHandler()

        if str(args.status) == 'active':
            if str(args.role) == 'primary':
                print(args.node_name)
                handler.configure_as_active_primary(args.node_name)
            else:
                config = None
                if args.config_file_path is not None:
                    config = read_config_file (args.config_file_path)
                status = CSConstants.CS_NODE_ACTIVE
                validate_joincluster_inputs(args)
                handler.configure_as_active_secondary(config, args.node_name, status, args.fetchfreshbackup, False, args.primary_node_hostname, args.primary_node_ssh_key_path)
        else:
            config = None
            if args.config_file_path is not None:
                config = read_config_file (args.config_file_path)
            status = CSConstants.CS_NODE_PASSIVE
            validate_joincluster_inputs(args)
            handler.configure_as_passive_secondary(config, args.node_name, status, args.fetchfreshbackup, False, args.primary_node_hostname, args.primary_node_ssh_key_path)

        return

    if str(args.operation) == 'export-conf':

        handler = CyOpsHAHandler()
        handler.exportconf()
        return

    if str(args.operation) == 'whitelist':
        handler = CyOpsHAHandler()
        handler.whitelist(args.nodes)
        return
    
    if str(args.operation) == 'list-nodes':
        handler = CyOpsHAHandler()
        handler.get_node_list(args.is_active, args.is_primary)
        return

    if str(args.operation) == 'takeover':

        handler = CyOpsHAHandler()
        handler.takeover_as_active_primary(interactive=args.interactive)
        return
  
    if str(args.operation) == 'restore':

        handler = CyOpsHAHandler()
        handler.restore_after_firedrill()
        return

    if str(args.operation) == 'firedrill':

        handler = CyOpsHAHandler()
        handler.firedrill()
        return 

    if str(args.operation) == 'leave-cluster':

        handler = CyOpsHAHandler()
        handler.leavecluster(args.interactive)
        return

    if str(args.operation) == 'set-node-name':
        handler = CyOpsHAHandler()
        handler.update_node_name(args.node_name)
        return

    return

if __name__ == '__main__':
    start()
